/* sane - Scanner Access Now Easy.
   Copyright (C) 1996, 1997 David Mosberger-Tang
   This file is part of the SANE package.

   SANE is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   SANE is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with sane; see the file COPYING.  If not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   This file provides a generic SCSI interface.  */

#include <sane/config.h>

#include <lalloca.h>	/* MUST come first for AIX! */

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/param.h>

#ifdef HAVE_SCSI_SG_H
# include <scsi/sg.h>
#else
# ifdef HAVE_SYS_SCSI_H
#  include <sys/scsi.h>
# else
#  ifdef HAVE_SYS_SCSIIO_H
#   include <sys/scsiio.h>
#   ifdef HAVE_SCSI_H
#    include <scsi.h>
#    define scsireq_t struct scsireq_t
#   endif
#  endif
# endif
#endif

#include <sane/sanei_scsi.h>
#include <sane/sanei.h>

#ifdef SG_BIG_BUFF
int sanei_scsi_max_request_size = SG_BIG_BUFF;
#else
/* FIXME: what to put here? */
int sanei_scsi_max_request_size = 512;
#endif

#define BACKEND_NAME	sanei_scsi
#include <sane/sanei_debug.h>

SANE_Status
sanei_scsi_open (const char * dev, int * fdp)
{
  int fd;

  DBG_INIT();

  fd = open(dev, O_RDWR | O_EXCL);
  if (fd < 0)
    return SANE_STATUS_INVAL;

#ifdef SG_SET_TIMEOUT
  /* Set large timeout since some scanners are slow but do not
     disconnect... ;-( */
  {
    int timeout;
    timeout = 10*60*HZ;		/* how about 10 minutes? ;-) */
    ioctl(fd, SG_SET_TIMEOUT, &timeout);
  }
#endif

  if (fdp)
    *fdp = fd;
  return SANE_STATUS_GOOD;
}


#ifdef HAVE_SCSI_SG_H
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  /* The Linux generic scsi interface is braindead as it forces us to
     copy things around like mad.  Fortunately, most command requests
     are small.  */
  struct
    {
      struct sg_header	hdr;
      u_int8_t		data[SG_BIG_BUFF];
    }
  *req;
  ssize_t nread;

  req = alloca (sizeof (*req));	/* works around gdb bug... ;-( */

  memset (&req->hdr, 0, sizeof (req->hdr));
  req->hdr.pack_len  = src_size + sizeof (req->hdr);
  req->hdr.reply_len = (dst_size ? *dst_size : 0) + sizeof (req->hdr);
  memcpy (&req->data, src, src_size);

  if (write (fd, req, req->hdr.pack_len) != req->hdr.pack_len)
    {
      DBG(1, "sanei_scsi_cmd: bad write\n");
      return SANE_STATUS_IO_ERROR;
    }

  nread = read (fd, req, req->hdr.reply_len);
  if (nread < 0)
    {
      DBG(1, "sanei_scsi_cmd: read returned %ld (errno=%d)\n",
	  (long) nread, errno);
      return SANE_STATUS_IO_ERROR;
    }

  nread -= sizeof (req->hdr);

  if (req->hdr.result != 0)
    {
      DBG(1, "sanei_scsi_cmd: SCSI completed with result=%s\n",
	  strerror (req->hdr.result));
      if (req->hdr.result == EBUSY)
	return SANE_STATUS_DEVICE_BUSY;
      else
	return SANE_STATUS_IO_ERROR;
    }

  if (dst)
    memcpy(dst, req->data, nread);

  if (dst_size)
    *dst_size = nread;

  return SANE_STATUS_GOOD;
}
#else /* !HAVE_SCSI_SG_H */

# ifdef HAVE_SYS_SCSIIO_H
/* XXX untested code! */

#  ifndef HAVE_SCSIREQ_ENTER
static int
scsireq_enter (int fd, scsireq_t *hdr)
{
  return ioctl (fd, SCIOCCOMMAND, hdr);
}
#  endif /* !HAVE_SCSIREQ_ENTER */

SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  scsireq_t hdr;
  size_t cmd_len;

  cmd_len = ((u_char *) src)[0] == 0x55 ? 10 : 6;

  memset (&hdr, 0, sizeof (hdr));
  memcpy (hdr.cmd, src, cmd_len);
  src_size -= cmd_len;
  src = (char *) src + cmd_len;

  if (*dst_size)
    {
      hdr.flags = SCCMD_READ;
      hdr.databuf = dst;
    }
  else
    {
      hdr.flags = SCCMD_WRITE;
      hdr.databuf = src;
    }
  hdr.timeout = 60000;			/* 1 minute timeout */

  if (scsireq_enter (fd, &hdr) < 0)
    {
      DBG(1, "sanei_scsi_cmd: scsi_reqenter() failed: %s\n",
	  strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }

  if (dst_size)
    *dst_size = hdr.datalen_used;

  return SANE_STATUS_GOOD;
}
# else /* !HAVE_SYS_SCSIIO_H */
#  ifdef HAVE_SYS_SCSI_H

/* XXX untested code! */

SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  struct sctl_io hdr;

  memset (&hdr, 0, sizeof (hdr));
  memcpy (hdr.cdb, src, src_size);
  hdr.flags = (*dst_size > 0) ? SCTL_READ : 0;
  hdr.cdb_length = src_size;
  hdr.data = dst;
  hdr.data_length = *dst_size;
  hdr.max_msecs = 60000;	/* 1 minute timeout */
  if (ioctl (fd, SIOC_IO, &hdr) < 0)
    {
      DBG(1, "sanei_scsi_cmd: ioctl(SIOC_IO) failed: %s\n",
	  strerror (errno));
      return SANE_STATUS_IO_ERROR;
    }
  if (hdr.cdb_status)
    DBG(1, "sanei_scsi_cmd: SCSI completed with cdb_status=%d\n",
	hdr.cdb_status);
  *dst_size = hdr.data_xfer;
  return SANE_STATUS_GOOD;
}

#  else /* !HAVE_SYS_SCSI_H */

/* Stubbed SCSI command. */
SANE_Status
sanei_scsi_cmd (int fd, const void * src, size_t src_size,
		void * dst, size_t * dst_size)
{
  return SANE_STATUS_UNSUPPORTED;
}

#  endif /* !HAVE_SYS_SCSI_H */
# endif /* !HAVE_SYS_SCSIIO_H */
#endif /* !HAVE_SCSI_SG_H */


void
sanei_scsi_close (int fd)
{
  close(fd);
}
